;;;; mailbox-cursor-test.scm

(use mailbox)
(use srfi-18)

(declare
	(usual-integrations)
	(inline)
	(block)
	(unsafe) )

(define-macro (current-thread-name)
	`(thread-name (current-thread)) )

(define *critical-section* (make-mutex (gensym 'critical-section)))

(define-macro (critical-section . ?body)
  `(dynamic-wind
     (lambda () (mutex-lock! *critical-section*))
     (lambda () ,@?body)
     (lambda () (mutex-unlock! *critical-section*)) ) )

(define-macro (thread-labeled-print . ARGS)
	`(critical-section
		 (print "Thread " (current-thread-name) " - " ,@ARGS) ) )

(define-constant TERM-NUM 5)

(define mailbox-one (make-mailbox 'one))

(define writer-thread-one
	(make-thread
		(lambda ()
			(thread-labeled-print "Started!")
			(let loop ([cnt 0])
				(thread-sleep! 0.25)
				(mailbox-send! mailbox-one (cons (current-thread-name) cnt))
				(if (= cnt TERM-NUM)
					(mailbox-send! mailbox-one 'quit)
					(loop (add1 cnt))) ) )
		'writer-one) )

(define writer-thread-two
	(make-thread
		(lambda ()
			(thread-labeled-print "Started!")
			(let loop ([cnt 0])
				(thread-sleep! 0.25)
				(mailbox-send! mailbox-one (cons (current-thread-name) cnt))
				(if (= cnt TERM-NUM)
					(mailbox-send! mailbox-one 'quit)
					(loop (add1 cnt))) ) )
		'writer-two) )

(define reader-thread-one
	(make-thread
		(lambda ()
			(thread-labeled-print "Started!")
			(let ([mbc (make-mailbox-cursor mailbox-one)])
				(let loop ([msg (mailbox-cursor-next mbc)])
					(thread-labeled-print "Message: " msg)
					(unless (eq? 'quit msg)
						(when (and (even? (cdr msg)) (not (zero? (modulo (cdr msg) 5))))
							(thread-labeled-print "Removing Message: " msg)
							(mailbox-cursor-extract-and-rewind! mbc) )
						(loop (mailbox-cursor-next mbc)) ) ) ) )
		'reader-one) )

(define reader-thread-two
	(make-thread
		(lambda ()
			(thread-labeled-print "Started!")
			(let ([mbc (make-mailbox-cursor mailbox-one)])
				(let loop ([msg (mailbox-cursor-next mbc)])
					(thread-labeled-print "Message: " msg)
					(unless (eq? 'quit msg)
						(when (odd? (cdr msg))
							(thread-labeled-print "Removing Message: " msg)
							(mailbox-cursor-extract-and-rewind! mbc) )
						(loop (mailbox-cursor-next mbc)) ) ) ) )
		'reader-two) )

(thread-start! writer-thread-one)
(thread-start! writer-thread-two)
(thread-start! reader-thread-one)
(thread-start! reader-thread-two)

(thread-join! writer-thread-one)
(thread-join! writer-thread-two)
(thread-join! reader-thread-one)
(thread-join! reader-thread-two)
