;;;; mailbox-timeout-test.scm

(use mailbox)
(use srfi-18)

(declare
	(usual-integrations)
	(inline)
	(block)
	(unsafe) )

(define-constant TERM-NUM 5)

(define-macro (current-thread-name)
	`(thread-name (current-thread)) )

(define-macro (current-seconds)
	`(time->seconds (current-time)) )

(define *critical-section* (make-mutex (gensym 'critical-section)))

(define-macro (critical-section . ?body)
  `(dynamic-wind
     (lambda () (mutex-lock! *critical-section*))
     (lambda () ,@?body)
     (lambda () (mutex-unlock! *critical-section*)) ) )

(define-macro (thread-labeled-print . ARGS)
	`(critical-section
		 (print "Thread " (current-thread-name) " - " ,@ARGS) ) )

(define mailbox-one (make-mailbox 'one))

(define writer-thread-one
	(make-thread
		(lambda ()
			(thread-labeled-print "Started!")
			(let loop ([cnt 0])
				(thread-labeled-print "Sleep @ " (current-seconds))
				(thread-sleep! 0.5 #;0.25)
				(thread-labeled-print "Awake @ " (current-seconds))
				(mailbox-send! mailbox-one (cons (current-thread-name) cnt))
				(if (= cnt TERM-NUM)
            (mailbox-send! mailbox-one 'quit)
            (loop (add1 cnt))) ) )
		'writer-one) )

(define reader-thread-one
	(make-thread
		(lambda ()
			(thread-labeled-print "Started!")
			(let loop ()
				(condition-case
						(begin
							(thread-labeled-print "Receiving @ " (current-seconds))
							(let ([msg (mailbox-receive! mailbox-one 0.25 #;0.5)])
								(thread-labeled-print " Message: " msg)
								(unless (eq? 'quit msg)
									(loop) ) ) )
					[(exn mailbox timeout)
						(thread-labeled-print "Timed Out @ " (current-seconds))
						(loop)]
					[exp ()
						(thread-labeled-print "Exception: " exp)]) ) )
		'reader-one) )

(thread-start! writer-thread-one)
(thread-start! reader-thread-one)

(thread-join! writer-thread-one)
(thread-join! reader-thread-one)
